@@ -6,6 +6,7 @@
 #include <cmath>
 #include <cstddef>
 #include <cstdint>
+#include <functional>
 #include <limits>
 #include <stdexcept>
 #include <string>
@@ -97,9 +98,8 @@ static std::vector<SourceDigit> binary_digits_to_greater_binary_base(
 
 template <class SourceDigit, class TargetDigit, std::size_t SOURCE_SHIFT,
           std::size_t TARGET_SHIFT,
-          std::size_t TARGET_DIGIT_MASK = power(TargetDigit(2), TARGET_SHIFT) -
-                                          1>
-static std::vector<SourceDigit> binary_digits_to_lesser_binary_base(
+          std::size_t TARGET_DIGIT_MASK = (TargetDigit(1) << TARGET_SHIFT) - 1>
+static std::vector<TargetDigit> binary_digits_to_lesser_binary_base(
     const std::vector<SourceDigit>& source) {
   static_assert(SOURCE_SHIFT > TARGET_SHIFT,
                 "Target base should be lesser than a source one.");
@@ -109,10 +109,10 @@ static std::vector<SourceDigit> binary_digits_to_lesser_binary_base(
       (result_digits_bits_count + (TARGET_SHIFT - 1)) / TARGET_SHIFT);
   std::vector<TargetDigit> result;
   result.reserve(result_digits_count);
-  double_precision_t<TargetDigit> accumulator = 0;
+  double_precision_t<SourceDigit> accumulator = 0;
   std::size_t accumulator_bits_count = 0;
   for (std::size_t index = 0; index < source.size(); ++index) {
-    accumulator |= static_cast<double_precision_t<TargetDigit>>(source[index])
+    accumulator |= static_cast<double_precision_t<SourceDigit>>(source[index])
                    << accumulator_bits_count;
     accumulator_bits_count += SOURCE_SHIFT;
     do {
@@ -238,6 +238,7 @@ class BigInt {
       _sign = 0;
       _digits = {0};
     } else {
+      _sign = 1;
       Digit remainder = value >> BINARY_SHIFT;
       if (remainder) {
         _digits.push_back(value & BINARY_DIGIT_MASK);
@@ -370,9 +371,6 @@ class BigInt {
   BigInt gcd(const BigInt& other) const {
     std::vector<Digit> largest_digits = _digits,
                        smallest_digits = other._digits;
-    if (largest_digits.size() <= 2 && smallest_digits.size() <= 2)
-      return BigInt(euclidean_algorithm(reduce_digits(largest_digits),
-                                        reduce_digits(smallest_digits)));
     if (digits_lesser_than(largest_digits, smallest_digits))
       std::swap(largest_digits, smallest_digits);
     std::size_t largest_digits_count, smallest_digits_count;
@@ -479,54 +477,37 @@ class BigInt {
       largest_digits = next_largest_digits;
       smallest_digits = next_smallest_digits;
     }
-    return BigInt(euclidean_algorithm(reduce_digits(largest_digits),
-                                      reduce_digits(smallest_digits)));
+    return BigInt(to_gcd(reduce_digits<DoubleDigit>(largest_digits),
+                         reduce_digits<DoubleDigit>(smallest_digits)));
   }
 
   void divmod(const BigInt& divisor, BigInt& quotient,
               BigInt& remainder) const {
-    std::size_t digits_count = _digits.size(),
-                divisor_digits_count = divisor._digits.size();
-    if (divisor._sign == 0)
-      throw std::range_error("Division by zero is undefined.");
-    else if (_sign == 0) {
-      quotient = BigInt();
-      remainder = *this;
-    } else if (digits_count < divisor_digits_count ||
-               (digits_count == divisor_digits_count &&
-                _digits.back() < divisor._digits.back())) {
-      if (_sign != divisor._sign) {
-        quotient = BigInt(-1, std::vector<Digit>({1}));
-        remainder = *this + divisor;
-      } else {
-        quotient = BigInt();
-        remainder = *this;
-      }
-    } else {
-      if (divisor_digits_count == 1) {
-        std::vector<Digit> quotient_digits;
-        Digit remainder_digit = divmod_digits_by_digit(
-            _digits, divisor._digits[0], quotient_digits);
-        quotient = BigInt(_sign * divisor._sign, quotient_digits);
-        remainder = BigInt(_sign * static_cast<SignedDigit>(remainder_digit));
-      } else {
-        std::vector<Digit> quotient_digits, remainder_digits;
-        divmod_two_or_more_digits(_digits, divisor._digits, quotient_digits,
-                                  remainder_digits);
-        quotient =
-            BigInt(_sign * divisor._sign *
-                       (quotient_digits.size() > 1 || quotient_digits[0] != 0),
-                   quotient_digits);
-        remainder = BigInt(
-            _sign * (remainder_digits.size() > 1 || remainder_digits[0] != 0),
-            remainder_digits);
-      }
-      if ((divisor._sign < 0 && remainder._sign > 0) ||
-          (divisor._sign > 0 && remainder._sign < 0)) {
-        quotient = quotient - BigInt(1, std::vector<Digit>({1}));
-        remainder = remainder + divisor;
-      }
+    divmod<true, true>(divisor, &quotient, &remainder);
+  }
+
+  BigInt floor_divide(const BigInt& divisor) const {
+    BigInt result;
+    divmod<true, false>(divisor, &result, nullptr);
+    return result;
+  }
+
+  BigInt invmod(const BigInt& divisor) const {
+    BigInt candidate, result{1u}, step_dividend = *this, step_divisor = divisor;
+    while (step_divisor) {
+      BigInt quotient, remainder;
+      step_dividend.divmod(step_divisor, quotient, remainder);
+      step_dividend = step_divisor;
+      step_divisor = remainder;
+      const BigInt next_candidate = result - quotient * candidate;
+      result = candidate;
+      candidate = next_candidate;
     }
+    if (step_dividend.is_one()) {
+      if (result.sign() < 0) result = result % divisor;
+      return result;
+    }
+    throw std::invalid_argument("Not invertible.");
   }
 
   BigInt operator+(const BigInt& other) const {
@@ -616,8 +597,209 @@ class BigInt {
                        : digits_lesser_than_or_equal(other._digits, _digits)));
   }
 
+  BigInt operator%(const BigInt& divisor) const {
+    BigInt result;
+    divmod<false, true>(divisor, nullptr, &result);
+    return result;
+  }
+
+  double operator/(const BigInt& divisor) const {
+    if (divisor._sign == 0)
+      throw std::range_error("Division by zero is undefined.");
+    bool negate = (_sign < 0) ^ (divisor._sign < 0);
+    if (sign() == 0) return negate ? -0.0 : 0.0;
+    const std::vector<Digit>& dividend_digits = digits();
+    const std::vector<Digit>& divisor_digits = divisor.digits();
+    std::size_t dividend_digits_count = dividend_digits.size();
+    std::size_t divisor_digits_count = divisor_digits.size();
+    bool dividend_is_small =
+        dividend_digits_count <= MANTISSA_BINARY_DIGITS_COUNT ||
+        (dividend_digits_count == MANTISSA_BINARY_DIGITS_COUNT + 1 &&
+         (dividend_digits[MANTISSA_BINARY_DIGITS_COUNT] >>
+          MANTISSA_EXTRA_BITS) == 0);
+    bool divisor_is_small =
+        divisor_digits_count <= MANTISSA_BINARY_DIGITS_COUNT ||
+        (divisor_digits_count == MANTISSA_BINARY_DIGITS_COUNT + 1 &&
+         (divisor_digits[MANTISSA_BINARY_DIGITS_COUNT] >>
+          MANTISSA_EXTRA_BITS) == 0);
+    if (dividend_is_small && divisor_is_small) {
+      double reduced_dividend = reduce_digits<double>(dividend_digits);
+      double reduced_divisor = reduce_digits<double>(divisor_digits);
+      double result = reduced_dividend / reduced_divisor;
+      return negate ? -result : result;
+    }
+    std::make_signed_t<std::size_t> digits_count_difference =
+        dividend_digits_count - divisor_digits_count;
+    if (digits_count_difference >
+        static_cast<std::make_signed_t<std::size_t>>(
+            std::numeric_limits<std::size_t>::max() / BINARY_SHIFT) -
+            1)
+      throw std::overflow_error(
+          "Division result too large to be expressed as floating point.");
+    else if (digits_count_difference <
+             1 - static_cast<std::make_signed_t<std::size_t>>(
+                     std::numeric_limits<std::size_t>::max() / BINARY_SHIFT))
+      return negate ? -0.0 : 0.0;
+    std::make_signed_t<std::size_t> bit_lengths_difference =
+        digits_count_difference * BINARY_SHIFT +
+        (to_bit_length(dividend_digits.back()) -
+         to_bit_length(divisor_digits.back()));
+    if (bit_lengths_difference > std::numeric_limits<double>::max_exponent)
+      throw std::overflow_error(
+          "Division result too large to be expressed as floating point.");
+    else if (bit_lengths_difference <
+             static_cast<std::make_signed_t<std::size_t>>(
+                 std::numeric_limits<double>::min_exponent - MANTISSA_BITS - 1))
+      return negate ? -0.0 : 0.0;
+    std::make_signed_t<std::size_t> shift =
+        std::max(bit_lengths_difference,
+                 static_cast<std::make_signed_t<std::size_t>>(
+                     std::numeric_limits<double>::min_exponent)) -
+        MANTISSA_BITS - 2;
+    bool inexact = false;
+    Digit* quotient_data;
+    std::size_t quotient_digits_count;
+    if (shift <= 0) {
+      std::size_t shift_digits = (-shift) / BINARY_SHIFT;
+      if (dividend_digits_count >=
+          std::numeric_limits<std::make_signed_t<std::size_t>>::max() - 1 -
+              shift_digits)
+        throw std::overflow_error(
+            "Division result too large to be expressed as floating point.");
+      quotient_digits_count = dividend_digits_count + shift_digits + 1;
+      quotient_data = new Digit[quotient_digits_count]();
+      Digit remainder = shift_digits_left(
+          dividend_digits.data(), dividend_digits_count,
+          (-shift) % BINARY_SHIFT, quotient_data + shift_digits);
+      quotient_data[dividend_digits_count + shift_digits] = remainder;
+    } else {
+      std::size_t shift_digits = shift / BINARY_SHIFT;
+      quotient_digits_count = dividend_digits_count - shift_digits;
+      quotient_data = new Digit[quotient_digits_count]();
+      Digit remainder = shift_digits_right(
+          dividend_digits.data() + shift_digits, quotient_digits_count,
+          shift % BINARY_SHIFT, quotient_data);
+      if (remainder) inexact = true;
+      while (!inexact && shift_digits > 0)
+        if (dividend_digits[--shift_digits]) inexact = true;
+    }
+    std::vector<Digit> quotient_digits(quotient_data,
+                                       quotient_data + quotient_digits_count);
+    normalize_digits(quotient_digits);
+    if (divisor_digits_count == 1) {
+      std::vector<Digit> next_quotient_digits;
+      Digit remainder = divmod_digits_by_digit(
+          quotient_digits, divisor_digits[0], next_quotient_digits);
+      std::swap(quotient_digits, next_quotient_digits);
+      if (remainder) inexact = true;
+    } else {
+      std::vector<Digit> next_quotient_digits, remainder;
+      divmod_two_or_more_digits(quotient_digits, divisor_digits,
+                                next_quotient_digits, remainder);
+      std::swap(quotient_digits, next_quotient_digits);
+      if (remainder.size() > 1 || remainder[0] != 0) inexact = true;
+    }
+    std::make_signed_t<std::size_t> quotient_bit_length =
+        (quotient_digits.size() - 1) * BINARY_SHIFT +
+        to_bit_length(quotient_digits.back());
+    std::make_signed_t<std::size_t> extra_bits =
+        std::max(quotient_bit_length,
+                 std::numeric_limits<double>::min_exponent - shift) -
+        MANTISSA_BITS;
+    const Digit mask = static_cast<Digit>(1) << (extra_bits - 1);
+    Digit quotient_low_digit = quotient_digits[0] | inexact;
+    if ((quotient_low_digit & mask) && (quotient_low_digit & (3U * mask - 1U)))
+      quotient_low_digit += mask;
+    quotient_digits[0] = quotient_low_digit & ~(2U * mask - 1U);
+    double reduced_quotient = reduce_digits<double>(quotient_digits);
+    if (shift + quotient_bit_length >=
+            std::numeric_limits<double>::max_exponent &&
+        (shift + quotient_bit_length >
+             std::numeric_limits<double>::max_exponent ||
+         reduced_quotient == ldexp(1.0, static_cast<int>(quotient_bit_length))))
+      throw std::overflow_error(
+          "Division result too large to be expressed as floating point.");
+    double result = ldexp(reduced_quotient, static_cast<int>(shift));
+    return negate ? -result : result;
+  }
+
   BigInt abs() const { return _sign < 0 ? BigInt(1, _digits) : *this; }
 
+  BigInt pow(BigInt exponent, const BigInt* maybe_modulus = nullptr) const {
+    BigInt base = *this, modulus;
+    bool is_negative = false;
+    std::function<BigInt(const BigInt&, const BigInt&)> make_step;
+    if (maybe_modulus != nullptr) {
+      if (maybe_modulus->sign() == 0)
+        throw std::invalid_argument("Modulus cannot be zero.");
+      is_negative = maybe_modulus->sign() < 0;
+      modulus = maybe_modulus->abs();
+      if (modulus.is_one()) return BigInt();
+      if (exponent.sign() < 0) {
+        exponent = -exponent;
+        base = base.invmod(modulus);
+      }
+      if (base.sign() < 0 || base.digits().size() > modulus.digits().size())
+        base = base % modulus;
+      make_step = [&modulus](const BigInt& first, const BigInt& second) {
+        return (first * second) % modulus;
+      };
+    } else if (exponent.sign() < 0)
+      throw std::range_error(
+          "Exponent should be positive or modulus should be specified.");
+    else
+      make_step = [](const BigInt& first, const BigInt& second) {
+        return first * second;
+      };
+    const std::vector<Digit>& exponent_digits = exponent.digits();
+    Digit exponent_digit = exponent_digits.back();
+    std::size_t exponent_digits_count = exponent_digits.size();
+    BigInt result = BigInt(1u);
+    if (exponent_digits_count == 1 && exponent_digit <= 3) {
+      if (exponent_digit >= 2) {
+        result = make_step(base, base);
+        if (exponent_digit == 3) result = make_step(result, base);
+      } else if (exponent_digit == 1)
+        result = make_step(base, result);
+    } else if (exponent_digits_count <= WINDOW_CUTOFF) {
+      result = base;
+      Digit bit = 2;
+      for (;; bit <<= 1)
+        if (bit > exponent_digit) {
+          bit >>= 1;
+          break;
+        }
+      bit >>= 1;
+      for (auto exponent_digit_position = exponent_digits.rbegin();;) {
+        for (; bit != 0; bit >>= 1) {
+          result = make_step(result, result);
+          if (exponent_digit & bit) result = make_step(result, base);
+        }
+        if (++exponent_digit_position == exponent_digits.rend()) break;
+        exponent_digit = *exponent_digit_position;
+        bit = static_cast<Digit>(1) << (BINARY_SHIFT - 1);
+      }
+    } else {
+      BigInt cache[WINDOW_BASE];
+      cache[0] = result;
+      for (std::size_t index = 1; index < WINDOW_BASE; ++index)
+        cache[index] = make_step(cache[index - 1], base);
+      std::vector<WindowDigit> exponent_window_digits =
+          binary_digits_to_binary_base<Digit, WindowDigit, BINARY_SHIFT,
+                                       WINDOW_SHIFT>(exponent_digits);
+      for (auto exponent_digit_position = exponent_window_digits.rbegin();
+           exponent_digit_position != exponent_window_digits.rend();
+           ++exponent_digit_position) {
+        const WindowDigit digit = *exponent_digit_position;
+        for (std::size_t iteration = 0; iteration < WINDOW_SHIFT; ++iteration)
+          result = make_step(result, result);
+        if (digit) result = make_step(result, cache[digit]);
+      }
+    }
+    if (is_negative && result) result = result - modulus;
+    return result;
+  }
+
   template <std::size_t BASE = 10,
             std::size_t TARGET_SHIFT = floor_log<BASE>(BINARY_BASE),
             std::size_t TARGET_BASE = power(BASE, TARGET_SHIFT)>
@@ -644,7 +826,9 @@ class BigInt {
       *--stop = '0';
     else if (_sign < 0)
       *--stop = '-';
-    return std::string(characters, characters_count);
+    std::string result(characters, characters_count);
+    delete[] characters;
+    return result;
   }
 
  protected:
@@ -653,6 +837,10 @@ class BigInt {
 
   const std::vector<Digit>& digits() const { return _digits; }
 
+  bool is_one() const {
+    return _sign > 0 && _digits.size() == 1 && _digits[0] == 1;
+  }
+
   int sign() const { return _sign; }
 
   SignedDigit signed_digit() const {
@@ -667,6 +855,24 @@ class BigInt {
   int _sign;
   std::vector<Digit> _digits;
 
+  static constexpr Digit KARATSUBA_CUTOFF = 70;
+  static constexpr Digit KARATSUBA_SQUARE_CUTOFF = KARATSUBA_CUTOFF * 2;
+  static constexpr std::size_t MANTISSA_BITS =
+      std::numeric_limits<double>::digits;
+  static constexpr double MANTISSA_BITS_POWER_OF_TWO =
+      power(2.0, MANTISSA_BITS);
+  static constexpr std::size_t MANTISSA_BINARY_DIGITS_COUNT =
+      MANTISSA_BITS / BINARY_SHIFT;
+  static constexpr std::make_signed_t<std::size_t> MANTISSA_EXTRA_BITS =
+      MANTISSA_BITS % BINARY_SHIFT;
+  static constexpr std::size_t WINDOW_CUTOFF = 8;
+  static constexpr std::size_t WINDOW_SHIFT = 5;
+  static constexpr std::size_t WINDOW_BASE = 1 << WINDOW_SHIFT;
+  static constexpr std::size_t WINDOW_DIGIT_MASK = WINDOW_BASE - 1;
+  using WindowDigit = std::uint8_t;
+  static_assert(WINDOW_SHIFT <= std::numeric_limits<WindowDigit>::digits,
+                "Window digit type should be able to contain window digits.");
+
   static void divmod_two_or_more_digits(const std::vector<Digit>& dividend,
                                         const std::vector<Digit>& divisor,
                                         std::vector<Digit>& quotient,
@@ -738,11 +944,14 @@ class BigInt {
     quotient = quotient_size ? std::vector<Digit>(quotient_data,
                                                   quotient_data + quotient_size)
                              : std::vector<Digit>({0});
+    delete[] quotient_data;
     normalize_digits(quotient);
     shift_digits_right(dividend_normalized, divisor_digits_count, shift,
                        divisor_normalized);
+    delete[] dividend_normalized;
     remainder = std::vector<Digit>(divisor_normalized,
                                    divisor_normalized + divisor_digits_count);
+    delete[] divisor_normalized;
     normalize_digits(remainder);
   }
 
@@ -751,7 +960,7 @@ class BigInt {
                                       std::vector<Digit>& quotient) {
     DoubleDigit remainder = 0;
     std::size_t digits_count = dividend.size();
-    Digit* quotient_data = new Digit[digits_count]();
+    Digit* const quotient_data = new Digit[digits_count]();
     for (std::size_t offset = 1; offset <= digits_count; ++offset) {
       remainder = (remainder << BINARY_SHIFT) | dividend[digits_count - offset];
       Digit quotient_digit = quotient_data[digits_count - offset] =
@@ -760,17 +969,19 @@ class BigInt {
     }
     quotient =
         std::vector<Digit>(quotient_data, quotient_data + dividend.size());
+    delete[] quotient_data;
     normalize_digits(quotient);
     return static_cast<Digit>(remainder);
   }
 
-  static constexpr Digit KARATSUBA_CUTOFF = 70;
-  static constexpr Digit KARATSUBA_SQUARE_CUTOFF = KARATSUBA_CUTOFF * 2;
-
-  static DoubleDigit reduce_digits(const std::vector<Digit>& digits) {
-    DoubleDigit result = 0;
+  template <class Result>
+  static Result reduce_digits(const std::vector<Digit>& digits) {
+    Result result = 0;
     for (auto position = digits.rbegin(); position != digits.rend(); ++position)
-      result = (result << BINARY_SHIFT) | *position;
+      if constexpr (std::is_integral<Result>())
+        result = (result << BINARY_SHIFT) | *position;
+      else
+        result = (result * BINARY_BASE) + *position;
     return result;
   }
 
@@ -1025,6 +1236,64 @@ class BigInt {
     if (digits_count != digits.size()) digits.resize(digits_count);
   }
 
+  template <bool WITH_QUOTIENT, bool WITH_REMAINDER>
+  void divmod(const BigInt& divisor, BigInt* quotient,
+              BigInt* remainder) const {
+    static_assert(WITH_QUOTIENT || WITH_REMAINDER,
+                  "Quotient or remainder or both should be requested.");
+    std::size_t digits_count = _digits.size(),
+                divisor_digits_count = divisor._digits.size();
+    if (divisor._sign == 0)
+      throw std::range_error("Division by zero is undefined.");
+    else if (_sign == 0) {
+      if constexpr (WITH_QUOTIENT) *quotient = BigInt();
+      if constexpr (WITH_REMAINDER) *remainder = *this;
+    } else if (digits_count < divisor_digits_count ||
+               (digits_count == divisor_digits_count &&
+                _digits.back() < divisor._digits.back())) {
+      if (_sign != divisor._sign) {
+        if constexpr (WITH_QUOTIENT)
+          *quotient = BigInt(-1, std::vector<Digit>({1}));
+        if constexpr (WITH_REMAINDER) *remainder = *this + divisor;
+      } else {
+        if constexpr (WITH_QUOTIENT) *quotient = BigInt();
+        if constexpr (WITH_REMAINDER) *remainder = *this;
+      }
+    } else {
+      int remainder_sign = _sign;
+      if (divisor_digits_count == 1) {
+        std::vector<Digit> quotient_digits;
+        Digit remainder_digit = divmod_digits_by_digit(
+            _digits, divisor._digits[0], quotient_digits);
+        remainder_sign *= remainder_digit != 0;
+        if constexpr (WITH_QUOTIENT)
+          *quotient = BigInt(_sign * divisor._sign, quotient_digits);
+        if constexpr (WITH_REMAINDER)
+          *remainder =
+              BigInt(_sign * static_cast<SignedDigit>(remainder_digit));
+      } else {
+        std::vector<Digit> quotient_digits, remainder_digits;
+        divmod_two_or_more_digits(_digits, divisor._digits, quotient_digits,
+                                  remainder_digits);
+        remainder_sign *=
+            remainder_digits.size() > 1 || remainder_digits[0] != 0;
+        if constexpr (WITH_QUOTIENT)
+          *quotient = BigInt(
+              _sign * divisor._sign *
+                  (quotient_digits.size() > 1 || quotient_digits[0] != 0),
+              quotient_digits);
+        if constexpr (WITH_REMAINDER)
+          *remainder = BigInt(remainder_sign, remainder_digits);
+      }
+      if ((divisor._sign < 0 && remainder_sign > 0) ||
+          (divisor._sign > 0 && remainder_sign < 0)) {
+        if constexpr (WITH_QUOTIENT)
+          *quotient = *quotient - BigInt(1, std::vector<Digit>({1}));
+        if constexpr (WITH_REMAINDER) *remainder = *remainder + divisor;
+      }
+    }
+  }
+
   void parse_binary_base_digits(const char* start, const char* stop,
                                 std::size_t base, std::size_t digits_count) {
     const std::size_t bits_per_character = floor_log2(base);
@@ -1080,11 +1349,6 @@ class BigInt {
     return accumulator;
   }
 
-  static constexpr std::size_t MANTISSA_BITS =
-      std::numeric_limits<double>::digits;
-  static constexpr double MANTISSA_BITS_POWER_OF_TWO =
-      power(2.0, MANTISSA_BITS);
-
   double frexp(int& exponent) const {
     Digit result_digits[2 + (MANTISSA_BITS + 1) / BINARY_SHIFT] = {
         0,
