@@ -22,6 +22,7 @@ namespace py = pybind11;
 #define MODULE_NAME cppbuiltins
 #define C_STR_HELPER(a) #a
 #define C_STR(a) C_STR_HELPER(a)
+#define FRACTION_NAME "Fraction"
 #define INT_NAME "int"
 #define LIST_ITERATOR_NAME "list_iterator"
 #define LIST_NAME "list"
@@ -123,11 +124,9 @@ static std::vector<BaseInt::Digit> int_to_digits(const py::int_& value) {
   PyLongObject* ptr = (PyLongObject*)value.ptr();
   Py_ssize_t signed_size = Py_SIZE(ptr);
   std::size_t size = Py_ABS(signed_size) + (signed_size == 0);
-  std::vector<BaseInt::Digit> result =
-      binary_digits_to_binary_base<digit, BaseInt::Digit, PyLong_SHIFT,
-                                   BaseInt::BINARY_SHIFT>(
-          std::vector<digit>(ptr->ob_digit, ptr->ob_digit + size));
-  return result;
+  return binary_digits_to_binary_base<digit, BaseInt::Digit, PyLong_SHIFT,
+                                      BaseInt::BINARY_SHIFT>(
+      std::vector<digit>(ptr->ob_digit, ptr->ob_digit + size));
 }
 
 static const char* pystr_to_ascii_c_str(const py::str& string) {
@@ -151,16 +150,17 @@ class Int : public BaseInt {
   Int(const py::str& value, std::size_t base)
       : BaseInt(pystr_to_ascii_c_str(value), base) {}
 
-  static Int from_state(const py::int_& value) { return Int(value); }
-
-  static py::int_ to_state(const Int& value) {
-    return py::reinterpret_steal<py::int_>((PyObject*)value.as_PyLong());
+  operator py::int_() const {
+    return py::reinterpret_steal<py::int_>((PyObject*)as_PyLong());
   }
 
   Int operator%(const Int& divisor) const {
-    Int quotient, result;
-    divmod(divisor, quotient, result);
-    return result;
+    try {
+      return Int(BaseInt::operator%(divisor));
+    } catch (const std::range_error& exception) {
+      PyErr_SetString(PyExc_ZeroDivisionError, exception.what());
+      throw py::error_already_set();
+    }
   }
 
   const Int& operator+() const { return *this; }
@@ -181,6 +181,15 @@ class Int : public BaseInt {
     return Int(BaseInt::operator-(other));
   }
 
+  double operator/(const BigInt& divisor) const {
+    try {
+      return BaseInt::operator/(divisor);
+    } catch (const std::range_error& exception) {
+      PyErr_SetString(PyExc_ZeroDivisionError, exception.what());
+      throw py::error_already_set();
+    }
+  }
+
   Int abs() const { return Int(BaseInt::abs()); }
 
   PyLongObject* as_PyLong() const {
@@ -195,20 +204,24 @@ class Int : public BaseInt {
     return result;
   }
 
-  void divmod(const BigInt& divisor, BigInt& quotient,
-              BigInt& remainder) const {
+  py::tuple divmod(const Int& divisor) const {
+    Int quotient, remainder;
     try {
       BaseInt::divmod(divisor, quotient, remainder);
     } catch (const std::range_error& exception) {
       PyErr_SetString(PyExc_ZeroDivisionError, exception.what());
       throw py::error_already_set();
     }
+    return py::make_tuple(quotient, remainder);
   }
 
   Int floor_divide(const Int& divisor) const {
-    Int result, remainder;
-    divmod(divisor, result, remainder);
-    return result;
+    try {
+      return Int(BaseInt::floor_divide(divisor));
+    } catch (const std::range_error& exception) {
+      PyErr_SetString(PyExc_ZeroDivisionError, exception.what());
+      throw py::error_already_set();
+    }
   }
 
   Int gcd(const Int& other) const { return Int(BaseInt::gcd(other)); }
@@ -236,12 +249,271 @@ class Int : public BaseInt {
     result -= (result == std::numeric_limits<std::size_t>::max());
     return static_cast<Py_hash_t>(result);
   }
+
+  Int invmod(const Int& divisor) const { return Int(BaseInt::invmod(divisor)); }
+
+  bool is_one() const { return BaseInt::is_one(); }
+
+  py::object pow(const Int& exponent,
+                 const Int* maybe_modulus = nullptr) const {
+    if (maybe_modulus == nullptr && exponent.sign() < 0) {
+      PyObject* result = PyFloat_Type.tp_as_number->nb_power(
+          (PyObject*)as_PyLong(), (PyObject*)exponent.as_PyLong(), Py_None);
+      if (!result) throw py::error_already_set();
+      return py::reinterpret_steal<py::object>(result);
+    }
+    return py::cast(Int(BaseInt::pow(exponent, maybe_modulus)));
+  }
+
+  int sign() const { return BaseInt::sign(); }
 };
 
 static std::ostream& operator<<(std::ostream& stream, const Int& value) {
   return stream << C_STR(MODULE_NAME) "." INT_NAME "('" << value.repr() << "')";
 }
 
+static py::object pow(const py::float_& base, const py::float_& exponent) {
+  PyObject* result = PyNumber_Power(base.ptr(), exponent.ptr(), Py_None);
+  if (!result) throw py::error_already_set();
+  return py::reinterpret_steal<py::object>(result);
+}
+
+class Fraction {
+ public:
+  Fraction() : _numerator(Int()), _denominator(Int(1)) {}
+
+  Fraction(const Int& numerator, const Int& denominator = Int(1))
+      : Fraction(numerator, denominator, std::true_type{}) {}
+
+  Fraction operator+(const Fraction& other) const {
+    return Fraction(
+        _numerator * other._denominator + _denominator * other._numerator,
+        _denominator * other._denominator);
+  }
+
+  Fraction operator+(const Int& other) const {
+    return Fraction(_numerator + _denominator * other, _denominator);
+  }
+
+  operator bool() const { return bool(_numerator); }
+
+  operator double() const { return _numerator / _denominator; }
+
+  bool operator==(const Fraction& other) const {
+    return _numerator == other._numerator && _denominator == other._denominator;
+  }
+
+  bool operator<=(const Fraction& other) const {
+    return sign() < other.sign() ||
+           (sign() == other.sign() &&
+            _numerator * other._denominator <= _denominator * other._numerator);
+  }
+
+  bool operator<(const Fraction& other) const {
+    return sign() < other.sign() ||
+           (sign() == other.sign() &&
+            _numerator * other._denominator < _denominator * other._numerator);
+  }
+
+  Fraction operator%(const Fraction& other) const {
+    return Fraction(
+        (_numerator * other._denominator) % (other._numerator * _denominator),
+        _denominator * other._denominator);
+  }
+
+  Fraction operator%(const Int& other) const {
+    return Fraction(_numerator % (other * _denominator), _denominator);
+  }
+
+  Fraction operator*(const Fraction& other) const {
+    const Int numerator_other_denominator_gcd =
+        _numerator.gcd(other._denominator);
+    const Int other_numerator_denominator_gcd =
+        _denominator.gcd(other._numerator);
+    return Fraction(
+        _numerator.floor_divide(numerator_other_denominator_gcd) *
+            other._numerator.floor_divide(other_numerator_denominator_gcd),
+        _denominator.floor_divide(other_numerator_denominator_gcd) *
+            other._denominator.floor_divide(numerator_other_denominator_gcd));
+  }
+
+  Fraction operator*(const Int& other) const {
+    const Int denominator_other_gcd = _denominator.gcd(other);
+    return Fraction(_numerator * other.floor_divide(denominator_other_gcd),
+                    _denominator.floor_divide(denominator_other_gcd));
+  }
+
+  Fraction operator-() const {
+    return Fraction(-_numerator, _denominator, std::false_type{});
+  }
+
+  const Fraction& operator+() const { return *this; }
+
+  Fraction operator-(const Fraction& other) const {
+    return Fraction(
+        _numerator * other._denominator - _denominator * other._numerator,
+        _denominator * other._denominator);
+  }
+
+  Fraction operator-(const Int& other) const {
+    return Fraction(_numerator - _denominator * other, _denominator);
+  }
+
+  Fraction operator/(const Fraction& other) const {
+    const Int numerators_gcd = _numerator.gcd(other._numerator);
+    const Int denominators_gcd = _denominator.gcd(other._denominator);
+    return Fraction(_numerator.floor_divide(numerators_gcd) *
+                        other._denominator.floor_divide(denominators_gcd),
+                    other._numerator.floor_divide(numerators_gcd) *
+                        _denominator.floor_divide(denominators_gcd));
+  }
+
+  Fraction operator/(const Int& other) const {
+    const Int numerators_gcd = _numerator.gcd(other);
+    return Fraction(_numerator.floor_divide(numerators_gcd),
+                    other.floor_divide(numerators_gcd) * _denominator);
+  }
+
+  Int ceil() const { return -(-_numerator).floor_divide(_denominator); }
+
+  Int floor() const { return _numerator.floor_divide(_denominator); }
+
+  const Int& denominator() const { return _denominator; }
+
+  py::tuple divmod(const Fraction& divisor) const {
+    Int quotient = floor_divide(divisor);
+    Fraction remainder = operator%(divisor);
+    return py::make_tuple(quotient, remainder);
+  }
+
+  Int floor_divide(const Fraction& other) const {
+    return (_numerator * other._denominator)
+        .floor_divide(other._numerator * _denominator);
+  }
+
+  Int floor_divide(const Int& other) const {
+    return _numerator.floor_divide(other * _denominator);
+  }
+
+  Py_hash_t hash() const {
+    static const Int HASH_MODULUS{_PyHASH_MODULUS};
+    Py_hash_t result;
+    try {
+      result =
+          (Int(_numerator.abs().hash()) * _denominator.invmod(HASH_MODULUS))
+              .hash();
+    } catch (const std::invalid_argument&) {
+      result = _PyHASH_INF;
+    }
+    if (_numerator.sign() < 0) result = -result;
+    return result - (result == -1);
+  }
+
+  const Int& numerator() const { return _numerator; }
+
+  Fraction pow(const Int& exponent) const {
+    if (exponent.sign() < 0) {
+      if (_numerator.sign() == 0) {
+        PyErr_SetString(PyExc_ZeroDivisionError,
+                        "Denominator should not be zero.");
+        throw py::error_already_set();
+      }
+      Int exponent_modulus = -exponent;
+      return _numerator.sign() < 0
+                 ? Fraction(Int(static_cast<BaseInt>(-_denominator)
+                                    .pow(exponent_modulus)),
+                            Int(static_cast<BaseInt>(-_numerator)
+                                    .pow(exponent_modulus)),
+                            std::false_type{})
+                 : Fraction(Int(static_cast<BaseInt>(_denominator)
+                                    .pow(exponent_modulus)),
+                            Int(static_cast<BaseInt>(_numerator)
+                                    .pow(exponent_modulus)),
+                            std::false_type{});
+    }
+    return Fraction(Int(static_cast<BaseInt>(_numerator).pow(exponent)),
+                    Int(static_cast<BaseInt>(_denominator).pow(exponent)),
+                    std::false_type{});
+  }
+
+  py::object pow(const Fraction& exponent) const {
+    return exponent._denominator.is_one()
+               ? py::cast(pow(exponent._numerator))
+               : ::pow(py::float_{double{*this}}, py::float_{double{exponent}});
+  }
+
+  int sign() const { return _numerator.sign(); }
+
+  operator Int() const { return sign() < 0 ? ceil() : floor(); }
+
+ private:
+  Int _numerator, _denominator;
+
+  template <bool NORMALIZE>
+  Fraction(const Int& numerator, const Int& denominator,
+           std::bool_constant<NORMALIZE>)
+      : _numerator(numerator), _denominator(denominator) {
+    if constexpr (NORMALIZE) {
+      if (_denominator.sign() == 0) {
+        PyErr_SetString(PyExc_ZeroDivisionError,
+                        "Denominator should not be zero.");
+        throw py::error_already_set();
+      }
+      if (_denominator.sign() < 0) {
+        _numerator = -_numerator;
+        _denominator = -_denominator;
+      }
+      Int gcd = _numerator.gcd(_denominator);
+      _denominator = _denominator.floor_divide(gcd);
+      _numerator = _numerator.floor_divide(gcd);
+    }
+  }
+};
+
+static Fraction operator+(const Int& self, const Fraction& other) {
+  return Fraction(self * other.denominator() + other.numerator(),
+                  other.denominator());
+}
+
+static Fraction operator-(const Int& self, const Fraction& other) {
+  return Fraction(self * other.denominator() - other.numerator(),
+                  other.denominator());
+}
+
+static Fraction operator%(const Int& self, const Fraction& other) {
+  return Fraction((self * other.denominator()) % other.numerator(),
+                  other.denominator());
+}
+
+static Fraction operator*(const Int& self, const Fraction& other) {
+  const Int self_other_denominator_gcd = self.gcd(other.denominator());
+  return Fraction(
+      self.floor_divide(self_other_denominator_gcd) * other.numerator(),
+      other.denominator().floor_divide(self_other_denominator_gcd));
+}
+
+static Fraction operator/(const Int& self, const Fraction& other) {
+  const Int self_other_numerator_gcd = self.gcd(other.numerator());
+  return Fraction(
+      self.floor_divide(self_other_numerator_gcd) * other.denominator(),
+      other.numerator().floor_divide(self_other_numerator_gcd));
+}
+
+Int floor_divide(const Int& self, const Fraction& other) {
+  return (self * other.denominator()).floor_divide(other.numerator());
+}
+
+py::object pow(const Int& base, const Fraction& exponent) {
+  return exponent.denominator().is_one()
+             ? py::cast(Fraction(base).pow(exponent.numerator()))
+             : pow(py::float_(double{base}), py::float_(double{exponent}));
+}
+
+static std::ostream& operator<<(std::ostream& stream, const Fraction& value) {
+  return stream << C_STR(MODULE_NAME) "." FRACTION_NAME "(" << value.numerator()
+                << ", " << value.denominator() << ")";
+}
+
 template <class Iterable>
 IterableState iterable_to_state(const Iterable& self) {
   IterableState result;
@@ -1004,36 +1276,111 @@ PYBIND11_MODULE(MODULE_NAME, m) {
       .def(py::init<const py::str&, std::size_t>(), py::arg("string"),
            py::arg("base") = 10)
       .def(py::init<const py::int_&>(), py::arg("value"))
+      .def(py::self + py::self)
+      .def(py::self == py::self)
+      .def(~py::self)
+      .def(py::self <= py::self)
+      .def(py::self < py::self)
+      .def(py::self % py::self)
+      .def(py::self * py::self)
+      .def(-py::self)
+      .def(+py::self)
+      .def(py::self - py::self)
+      .def(py::self / py::self)
+      .def(py::pickle([](const Int& self) { return py::int_(self); },
+                      [](const py::int_& state) { return Int(state); }))
+      .def("__abs__", &Int::abs)
+      .def("__bool__", &Int::operator bool)
+      .def("__ceil__", &identity<const Int&>)
+      .def("__copy__", [](const Int& self) -> const Int& { return self; })
+      .def("__divmod__", &Int::divmod, py::is_operator{})
+      .def("__deepcopy__",
+           [](const Int& self, const py::dict&) -> Int { return self; })
+      .def("__float__", &Int::operator double)
+      .def("__floor__", &identity<const Int&>)
+      .def("__floordiv__", &Int::floor_divide, py::is_operator{})
+      .def("__hash__", &Int::hash)
+      .def("__int__", &Int::operator py::int_)
+      .def("__pow__", &Int::pow, py::arg("exponent"),
+           py::arg("modulus") = nullptr, py::is_operator{})
+      .def("__repr__", &to_repr<Int>)
+      .def("__str__", &Int::repr<10>)
+      .def("__trunc__", &identity<const Int&>);
+
+  py::class_<Fraction> PyFraction(m, FRACTION_NAME);
+  PyFraction.def(py::init<>())
+      .def(py::init<const Int&, const Int&>(), py::arg("numerator"),
+           py::arg("denominator") = Int(1))
+      .def(py::self + py::self)
+      .def(py::self + Int{})
       .def(py::self == py::self)
       .def(py::self <= py::self)
       .def(py::self < py::self)
-      .def(py::self + py::self)
       .def(py::self % py::self)
-      .def(~py::self)
-      .def(+py::self)
+      .def(py::self % Int{})
       .def(py::self * py::self)
+      .def(py::self * Int{})
       .def(-py::self)
+      .def(+py::self)
+      .def(Int{} + py::self)
+      .def(Int{} % py::self)
+      .def(Int{} * py::self)
+      .def(Int{} - py::self)
+      .def(Int{} / py::self)
       .def(py::self - py::self)
-      .def(py::pickle(&Int::to_state, &Int::from_state))
-      .def("__abs__", &Int::abs)
-      .def("__hash__", &Int::hash)
-      .def("__bool__", &Int::operator bool)
-      .def("__copy__", [](const Int& self) -> const Int& { return self; })
+      .def(py::self - Int{})
+      .def(py::self / py::self)
+      .def(py::self / Int{})
+      .def(py::pickle(
+          [](const Fraction& value) {
+            return py::make_tuple(value.numerator(), value.denominator());
+          },
+          [](const py::tuple& state) {
+            if (state.size() != 2) throw std::runtime_error("Invalid state.");
+            return Fraction(state[0].cast<Int>(), state[1].cast<Int>());
+          }))
+      .def("__bool__", &Fraction::operator bool)
+      .def("__ceil__", &Fraction::ceil)
+      .def("__divmod__", &Fraction::divmod, py::is_operator{})
+      .def("__float__", &Fraction::operator double)
+      .def("__floor__", &Fraction::floor)
+      .def("__floordiv__",
+           py::overload_cast<const Fraction&>(&Fraction::floor_divide,
+                                              py::const_),
+           py::is_operator{})
+      .def("__floordiv__",
+           py::overload_cast<const Int&>(&Fraction::floor_divide, py::const_),
+           py::is_operator{})
+      .def("__hash__", &Fraction::hash)
+      .def("__int__", [](const Fraction& self) { return py::int_(Int(self)); })
+      .def("__pow__", py::overload_cast<const Int&>(&Fraction::pow, py::const_),
+           py::arg("exponent"), py::is_operator{})
+      .def("__pow__",
+           py::overload_cast<const Fraction&>(&Fraction::pow, py::const_),
+           py::arg("exponent"), py::is_operator{})
+      .def("__repr__", &to_repr<Fraction>)
       .def(
-          "__divmod__",
-          [](const Int& self, const Int& other) {
-            Int quotient, remainder;
-            self.divmod(other, quotient, remainder);
-            return py::make_tuple(quotient, remainder);
+          "__rfloordiv__",
+          [](const Fraction& divisor, const Int& dividend) {
+            return floor_divide(dividend, divisor);
           },
           py::is_operator{})
-      .def("__deepcopy__",
-           [](const Int& self, const py::dict&) -> Int { return self; })
-      .def("__int__", &Int::to_state)
-      .def("__float__", &Int::operator double)
-      .def("__floordiv__", &Int::floor_divide, py::is_operator{})
-      .def("__repr__", &to_repr<Int>)
-      .def("__str__", [](const Int& self) { return self.repr(); });
+      .def(
+          "__rpow__",
+          [](const Fraction& exponent, const Int& base) {
+            return pow(base, exponent);
+          },
+          py::is_operator{})
+      .def("__str__",
+           [](const Fraction& self) {
+             return self.denominator().is_one()
+                        ? self.numerator().repr()
+                        : self.numerator().repr() + "/" +
+                              self.denominator().repr();
+           })
+      .def("__trunc__", &Fraction::operator Int)
+      .def_property_readonly("denominator", &Fraction::denominator)
+      .def_property_readonly("numerator", &Fraction::numerator);
 
   py::class_<List> PyList(m, LIST_NAME);
   PyList.def(py::init<py::iterable>(), py::arg("values"))
